#include "BuildingParameters.h"

/*
*		Definition of fixed coordinate used in this simulator:
*
*		Y=North
*		|
*		|
*		|
*		|------->X=East
*/

double BuildingParameters::getWindDirectionALaMathematician(double blowing_direction_measured_from_north_with_clockwise_in_degrees) const
{
	ML_TRACE("BuildingParameters::getWindDirectionALaMathematician");

	const double deg = blowing_direction_measured_from_north_with_clockwise_in_degrees;
	return DEG2RAD*(deg <= 90 ? 90 - deg : 450 - deg);
}

bool BuildingParameters::isTallStack() const
{
	ML_TRACE("BuildingParameters::isTallStack");

	// Placeholders for source object
	Ring _rin_; bool _stk_, _plm_; double _emi_,_vel_,_gas_,_tpr_;
	ptr_src->getPointSrc(_rin_, _stk_, _plm_, _emi_, _vel_, _gas_, _tpr_);

	// Placeholders for ring object
	int  _ide_; Vector3d _ctr_; double _dia_,_hhh_; string _des_;
	_rin_.getRing(_ide_, _ctr_, _dia_, _hhh_,_des_);

	return (_ctr_.Z() > gepStackHeight());
}

double BuildingParameters::distance(Tetragon ttr) const
{
	ML_TRACE("BuildingParameters::distance");

	// Placeholders for source object
	Ring _rin_; bool _stk_, _plm_; double _emi_,_vel_,_gas_,_tpr_;
	ptr_src->getPointSrc(_rin_, _stk_, _plm_, _emi_, _vel_, _gas_, _tpr_);

	// Placeholders for ring object
	int  _ide_; Vector3d _ctr_; double _dia_,_hhh_; string _des_;
	_rin_.getRing(_ide_, _ctr_, _dia_, _hhh_, _des_);

	Vector2d pt1,pt2,pt4,dpt; Rectangle3d rect;

	pt1 = ttr.pt(1);
	pt2 = ttr.pt(2);
	pt4 = ttr.pt(4);
	rect.Origin() = Vector3d(pt1.X(), pt1.Y(), 0.0);

	dpt = pt2 - pt1;
	rect.Edge0 () = Vector3d(dpt.X(), dpt.Y(), 0.0);
	dpt = pt4 - pt1;
	rect.Edge1 () = Vector3d(dpt.X(), dpt.Y(), 0.0);

	return Distance(Vector3d(_ctr_.X(), _ctr_.Y(), 0.0), rect);
}

void BuildingParameters::front()
{
	ML_TRACE("BuildingParameters::front");

	Tetragon ttr;
	Vector2d n12,n23,n34,n41;
	Vector2d N12,N23,N34,N41;
	double   c12,c23,c34,c41;


	// Evaluate single building direction ...


	ttr = ptr_blg->getTetragon(Building::SINGLE);

	// Define internal normal vectors
	n12 = Vector2d(ttr.pt(3) - ttr.pt(2));
	N12 = n12/n12.Length();
	n23 = Vector2d(ttr.pt(4) - ttr.pt(3));
	N23 = n23/n23.Length();
	n34 = Vector2d(ttr.pt(1) - ttr.pt(4));
	N34 = n34/n34.Length();
	n41 = Vector2d(ttr.pt(2) - ttr.pt(1));
	N41 = n41/n41.Length();

	// Evaluate directional cosines ...
	c12 = N12.Dot(e1);
	c23 = N23.Dot(e1);
	c34 = N34.Dot(e1);
	c41 = N41.Dot(e1);

	double cosine;

	if(false){/* Dummy */}
	else if(c12 >= 0.5*SQRT_2)
	{
		setFront(ttr.pt(1), ttr.pt(2), N12);
		setwl((ttr.pt(1) - ttr.pt(2)).Length(), (ttr.pt(2) - ttr.pt(3)).Length());
		cosine = c12;
	}
	else if(c23 >= 0.5*SQRT_2)
	{
		setFront(ttr.pt(2), ttr.pt(3), N23);
		setwl((ttr.pt(2) - ttr.pt(3)).Length(), (ttr.pt(3) - ttr.pt(4)).Length());
		cosine = c23;
	}
	else if(c34 >= 0.5*SQRT_2)
	{
		setFront(ttr.pt(3), ttr.pt(4), N34);
		setwl((ttr.pt(3) - ttr.pt(4)).Length(), (ttr.pt(4) - ttr.pt(1)).Length());
		cosine = c34;
	}
	else
	{
		setFront(ttr.pt(4), ttr.pt(1), N41);
		setwl((ttr.pt(4) - ttr.pt(1)).Length(), (ttr.pt(1) - ttr.pt(2)).Length());
		cosine = fabs(c41);
	}

	setTheta(acos(cosine));



	// Evaluate clusterized building direction ...

	// Placeholders for source object
	Ring _rin_; bool _stk_, _plm_; double _emi_,_vel_,_gas_,_tpr_;
	ptr_src->getPointSrc(_rin_, _stk_, _plm_, _emi_, _vel_, _gas_, _tpr_);

	// Placeholders for ring object
	int  _ide_; Vector3d _ctr_; double _dia_,_hhh_; string _des_;
	_rin_.getRing(_ide_, _ctr_, _dia_, _hhh_, _des_);

	const Vector2d ctr = Vector2d(_ctr_.X(), _ctr_.Y());

	ttr = ptr_blg->getTetragon(Building::CLUSTER);

	/*
	cout << ttr.pt(1).X() << ',' << ttr.pt(1).Y() << endl;
	cout << ttr.pt(2).X() << ',' << ttr.pt(2).Y() << endl;
	cout << ttr.pt(3).X() << ',' << ttr.pt(3).Y() << endl;
	cout << ttr.pt(4).X() << ',' << ttr.pt(4).Y() << endl;
	*/

	// Define internal normal vectors
	n12 = Vector2d(ttr.pt(3) - ttr.pt(2));
	N12 = n12/n12.Length();
	n23 = Vector2d(ttr.pt(4) - ttr.pt(3));
	N23 = n23/n23.Length();
	n34 = Vector2d(ttr.pt(1) - ttr.pt(4));
	N34 = n34/n34.Length();
	n41 = Vector2d(ttr.pt(2) - ttr.pt(1));
	N41 = n41/n41.Length();

	// Evaluate directional cosines ...
	c12 = N12.Dot(e1);
	c23 = N23.Dot(e1);
	c34 = N34.Dot(e1);
	c41 = N41.Dot(e1);

	if(false){/* Dummy */}
	else if(c12 >= 0.5*SQRT_2)
	{
		const Vector2d edge = (ttr.pt(1) - ttr.pt(2))/(ttr.pt(1) - ttr.pt(2)).Length();
		setWL(edge.Dot(ttr.pt(1) - ctr), edge.Dot(ttr.pt(2) - ctr), (ttr.pt(2) - ttr.pt(3)).Length());
	}
	else if(c23 >= 0.5*SQRT_2)
	{
		const Vector2d edge = (ttr.pt(2) - ttr.pt(3))/(ttr.pt(2) - ttr.pt(3)).Length();
		setWL(edge.Dot(ttr.pt(2) - ctr), edge.Dot(ttr.pt(3) - ctr), (ttr.pt(3) - ttr.pt(4)).Length());
	}
	else if(c34 >= 0.5*SQRT_2)
	{
		const Vector2d edge = (ttr.pt(3) - ttr.pt(4))/(ttr.pt(3) - ttr.pt(4)).Length();
		setWL(edge.Dot(ttr.pt(3) - ctr), edge.Dot(ttr.pt(4) - ctr), (ttr.pt(4) - ttr.pt(1)).Length());
	}
	else
	{
		const Vector2d edge = (ttr.pt(4) - ttr.pt(1))/(ttr.pt(4) - ttr.pt(1)).Length();
		setWL(edge.Dot(ttr.pt(4) - ctr), edge.Dot(ttr.pt(1) - ctr), (ttr.pt(1) - ttr.pt(2)).Length());
	}
}

double BuildingParameters::wb2() const
{
	// ML_TRACE("BuildingParameters::wb2");

	double _www_, _lll_; getwl(_www_, _lll_);

	return _www_*cos(getTheta()) + _lll_*sin(getTheta());
}

void BuildingParameters::lookupEffectiveBuilding()
{
	ML_TRACE("BuildingParameters::lookupEffectiveBuilding");

    e_f = 0; double maxgep = 0.0; double mindst = 1.0e10; ptr_blg->head();
    for(int i = 1; i<=ptr_blg->getNoElms(); i++)
	{
		front();	// First evaluate each building's wb1, wb2 and hb.

    	if(!isTallStack())
    	{
			const Tetragon blg_ttr = ptr_blg->getTetragon(Building::SINGLE); Tetragon src_ttr;

			// Placeholders for source object
			Ring _rin_; bool _stk_, _plm_; double _emi_,_vel_,_gas_,_tpr_;
			ptr_src->getPointSrc(_rin_, _stk_, _plm_, _emi_, _vel_, _gas_, _tpr_);

			// Placeholders for ring object
			int  _ide_; Vector3d _ctr_; double _dia_,_hhh_; string _des_;
			_rin_.getRing(_ide_, _ctr_, _dia_, _hhh_, _des_);

			const Vector2d ctr = Vector2d(_ctr_.X(), _ctr_.Y());

    		src_ttr.pt(1) = ctr - 5.0*l_star()*e1 - 1.0*l_star()*e2;
    		src_ttr.pt(2) = ctr + 5.0*l_star()*e1 - 1.0*l_star()*e2;
    		src_ttr.pt(3) = ctr + 5.0*l_star()*e1 + 1.0*l_star()*e2;
    		src_ttr.pt(4) = ctr - 5.0*l_star()*e1 + 1.0*l_star()*e2;

    		if(ptr_blg->distance(src_ttr) < l_star()*1.0e-2)
    		{
    			const double gep = gepStackHeight(); const double dst = distance(blg_ttr);
    			if(gep > maxgep || (fabs(gep - maxgep) < 1.0e-6 && dst < mindst))
				{
				    maxgep = gep;
				    mindst = dst;
				    e_f  = i;
				}
			}
		}
		ptr_blg->advance();
	}

	setMaxGEP(maxgep);
}

void BuildingParameters::exec()
{
	ML_TRACE("BuildingParameters::exec");

	CLST << endl;
	CLST << endl;
	CLST << "<< Evaluated building parameters list >>" << endl;
	CLST << endl;
	CLST << "+ MODEL-READY DIFFUSION PARAMETERS RELATED TO BUILDING DOWNWASH +" << endl;
	CLST << "^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^" << endl;
	CLST << setw(COLUMN) << "Stack IDs" << setw(COLUMN) << "Wind dir (1-16)" << setw(COLUMN) << "Bldg Width 1 (m)" << setw(COLUMN) << "Bldg Width 2 (m)" << setw(COLUMN) << "Bldg Height (m)" << setw(COLUMN) << "Cluster type (1-2)" << setw(COLUMN) << "Wake He (m)" << setw(COLUMN) << "Stack transl. x (m)" << setw(COLUMN) << "Stack transl. y (m)" << setw(COLUMN) << "Vert param (Cz1)" << setw(COLUMN) << "Vert param (Cz2)" << setw(COLUMN) << "Horiz param (Cy1)" << setw(COLUMN) << "Horiz param (Cy2)" << setw(COLUMN) << "Wake wind factor" << setw(COLUMN) << "Bldg IDs" << setw(COLUMN) << "Cross angle" << setw(COLUMN) << "GEP Height (m)" << endl;
	CLST << "----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------" << endl;

    for(int w_d = 1; w_d <= 16; w_d++)			// Loop for the 16 wind directions...
	{
		const double rad = getWindDirectionALaMathematician(22.5*(w_d + 8));
		// cout << w_d << endl;
		e1 = Vector2d( cos(rad), sin(rad));
		e2 = Vector2d(-sin(rad), cos(rad));

		lookupEffectiveBuilding(); evaluate(); dump(w_d);
	}
}

void BuildingParameters::evaluate()
{
	ML_TRACE("BuildingParameters::evaluate");

	// Placeholders for source object
	Ring _rin_; bool _stk_, _plm_; double _emi_,_vel_,_gas_,_tpr_;
	ptr_src->getPointSrc(_rin_, _stk_, _plm_, _emi_, _vel_, _gas_, _tpr_);

	// Placeholders for ring object
	int  _ide_; Vector3d _ctr_; double _dia_,_hhh_; string _des_;
	_rin_.getRing(_ide_, _ctr_, _dia_, _hhh_, _des_);

	// Initialize parameters ...
	h_h = w_1 = w_2 = h_e = d_x = d_y = z_1 = z_2 = y_1 = y_2 = a_l = 0.0; t_p = 0; h_e = _ctr_.Z();

	if(eff() == 0) return;

	ptr_blg->head(); for(int i = 1; i < eff(); i++) ptr_blg->advance(); front();

	h_h = hbl();
	w_1 = wb1();
	w_2 = wb2();
	h_e = hef();
	a_l = alp();
	t_p = typ();
	e_f = eff();
	getCs(z_1, z_2, y_1, y_2);
	getVs(d_x, d_y);
}

double BuildingParameters::hef() const
{
	ML_TRACE("BuildingParameters::hef");

	// Placeholders for source object
	Ring _rin_; bool _stk_, _plm_; double _emi_,_vel_,_gas_,_tpr_;
	ptr_src->getPointSrc(_rin_, _stk_, _plm_, _emi_, _vel_, _gas_, _tpr_);

	// Placeholders for ring object
	int  _ide_; Vector3d _ctr_; double _dia_,_hhh_; string _des_;
	_rin_.getRing(_ide_, _ctr_, _dia_, _hhh_, _des_);

	const double _h_s = _ctr_.Z();
	const double _h_b = hbl();
	const double _w_1 = wb1();
	const double _w_2 = wb2();
	const double nod1 = _w_1/_h_b;
	const double nod2 = _w_2/_h_b;
	const double nod3 = _h_s/_h_b;

	double res = _h_s;

	if(false){/* dummy */}
	else if(nod3 <= 0.5){res = 0.0;}
	else if(nod3 <= 1.0){res = 0.5*_h_s;}
	else if(nod3 <  2.5 && nod1 <= 1){res = _h_s*(typ() == 1 ? 0.56 : 0.67);}
	else if(nod3 <  2.5 && nod1 >  1){res = _h_s*0.44;}

	return res;
}

double BuildingParameters::alp() const
{
	ML_TRACE("BuildingParameters::alp");

	// Placeholders for source object
	Ring _rin_; bool _stk_, _plm_; double _emi_,_vel_,_gas_,_tpr_;
	ptr_src->getPointSrc(_rin_, _stk_, _plm_, _emi_, _vel_, _gas_, _tpr_);

	// Placeholders for ring object
	int  _ide_; Vector3d _ctr_; double _dia_,_hhh_; string _des_;
	_rin_.getRing(_ide_, _ctr_, _dia_, _hhh_, _des_);

	const double _h_s = _ctr_.Z();
	const double _h_b = hbl();
	const double _w_1 = wb1();
	const double _w_2 = wb2();
	const double nod1 = _w_1/_h_b;
	const double nod2 = _w_2/_h_b;
	const double nod3 = _h_s/_h_b;

	double res;

	if(false){/* dummy */}
	else if(nod3 > 1)
	{
		if(false){/* dummy */}
		else if(nod1 <  1){res = 0.76;}
		else if(nod1 <= 5){res = 0.8   - 0.039*nod1;}
		else              {res = 0.61;}
	}
	else if(typ() == 1)
	{
		if(false){/* dummy */}
		else if(nod1 <  1){res = 0.66;}
		else if(nod1 <= 5){res = 0.72  - 0.056*nod1;}
		else              {res = 0.44;}
	}
	else
	{
		if(false){/* dummy */}
		else if(nod1 <  1){res = 0.72;}
		else if(nod1 <= 3){res = 0.913 - 0.194*nod1;}
		else              {res = 0.33;}
	}

	if(nod1 > 1)
	{
		if(nod3 > 1)
		{
			res = res*(1 + 0.013*getTheta(true) - 2.2e-4*getTheta(true)*getTheta(true));
		}
		else
		{
			if(typ() == 1)
			{
				res = res*(1 + 0.053*getTheta(true) - 8.30e-4*getTheta(true)*getTheta(true));
			}
			else
			{
				res = res*(1 + 0.099*getTheta(true) - 1.98e-3*getTheta(true)*getTheta(true));
			}
		}
	}

	return res;
}

int BuildingParameters::typ() const
{
	// ML_TRACE("BuildingParameters::typ");

	double _WW1_, _WW2_, _LLL_; getWL(_WW1_, _WW2_, _LLL_);
	// cout << _WW1_ << ',' << _WW2_ << ',' << _LLL_ << endl;
	return ((_LLL_ >= 5*hbl() && fabs(_WW1_) >= 3*hbl() && fabs(_WW2_) >= 3*hbl() && _WW1_*_WW2_ < 0) ? 2 : 1);
}

void BuildingParameters::getCs(double& z_1_, double& z_2_, double& y_1_, double& y_2_) const
{
	ML_TRACE("BuildingParameters::getCs");

	// Placeholders for source object
	Ring _rin_; bool _stk_, _plm_; double _emi_,_vel_,_gas_,_tpr_;
	ptr_src->getPointSrc(_rin_, _stk_, _plm_, _emi_, _vel_, _gas_, _tpr_);

	// Placeholders for ring object
	int  _ide_; Vector3d _ctr_; double _dia_,_hhh_; string _des_;
	_rin_.getRing(_ide_, _ctr_, _dia_, _hhh_, _des_);

	const double _h_s = _ctr_.Z();
	const double _h_b = hbl();
	const double _w_1 = wb1();
	const double _w_2 = wb2();
	const double nod1 = _w_1/_h_b;
	const double nod2 = _w_2/_h_b;
	const double nod3 = _h_s/_h_b;
	double a, b;

	// Evaluate cz1
	if(false){/* dummy */}
	else if(nod1 <= 5)
	{
		a = (-0.00125*nod1 + 0.02 )*nod1 - 0.329;
		b = (-0.0045 *nod1 + 0.051)*nod1 + 0.645;
	}
	else
	{
		a = -0.26;
		b =  0.788;
	}
	z_1_ = a*nod3 + b;
	if(nod3 <= 1.0){z_1_ = z_1_*(typ() == 1 ? 1 - 0.010*getTheta(true) : 1 - 0.003*getTheta(true));}


	// Evaluate cz2
	if(false){/* dummy */}
	else if(typ() == 1)
	{
		a = 0.039;
		if(false){/* dummy */}
		else if(nod1 <  1){b = 0.0038;}
		else if(nod1 <= 5){b = 0.0137*nod1 - 0.0085;}
		else              {b = 0.059;}
		z_2_ = a*nod3 + b;
	}
	else
	{
		z_2_ = 0.052;
	}

	if(false){/* dummy */}
	else if(typ() == 1 && nod1 > 1)
	{
		if(false){/* dummy */}
		else if(nod3 >  1.0){z_2_ = z_2_*(1 - 0.0136*getTheta(true));}
		else                {z_2_ = z_2_*(1 - 0.0072*getTheta(true));}
	}
	else if(typ() == 2 && nod3 <= 1.0)
	{
		z_2_ = z_2_*(1 - 0.0098*getTheta(true));
	}


	// Evaluate cy1
	if(false){/* dummy */}
	else if(nod2 <= 5)
	{
		a = (-0.0170*nod2 + 0.173)*nod2 - 0.80;
		b = ( 0.0464*nod2 - 0.461)*nod2 + 1.93;
	}
	else
	{
		a = -0.36;
		b = 0.791;
	}
	y_1_ = a*nod3 + b;

	if(false){/* dummy */}
	else if(nod3 <= 1)
	{
		y_1_ = y_1_*(typ() == 1 ? 1 - 0.015*getTheta(true) : 1 - 0.0069*getTheta(true));
	}


	// Evaluate cy2
	y_2_ = 0.039;

	if(false){/* dummy */}
	else if(typ() == 2 && nod3 >  1)
	{
		y_2_ = y_2_*(1 + 0.0149*getTheta(true));
	}
	else if(typ() == 2 && nod3 <= 1)
	{
		y_2_ = y_2_*(1 + 0.0190*getTheta(true));
	}
}

void BuildingParameters::getVs(double& d_x_, double& d_y_) const
{
	ML_TRACE("BuildingParameters::getVs");

	d_x_ = d_y_ = 0.0;

	// Placeholders for source object
	Ring _rin_; bool _stk_, _plm_; double _emi_,_vel_,_gas_,_tpr_;
	ptr_src->getPointSrc(_rin_, _stk_, _plm_, _emi_, _vel_, _gas_, _tpr_);

	// Placeholders for ring object
	int  _ide_; Vector3d _ctr_; double _dia_,_hhh_; string _des_;
	_rin_.getRing(_ide_, _ctr_, _dia_, _hhh_, _des_);

	Vector2d _fr1_, _fr2_, _nrm_; getFront(_fr1_, _fr2_, _nrm_);
	const Vector3d _Fr1_ =  Vector3d(_fr1_.X(), _fr1_.Y(), 0.0);
	const Vector3d _Fr2_ =  Vector3d(_fr2_.X(), _fr2_.Y(), 0.0);
	Segment3d _Fr0_; _Fr0_.Origin() = _Fr1_; _Fr0_.Direction() =  _Fr2_ - _Fr1_;

	const Vector3d _Ctr_ =  Vector3d(_ctr_.X(), _ctr_.Y(), 0.0);

	const Tetragon  ttr = ptr_blg->getTetragon(Building::SINGLE);
	const Vector2d  pt1 = ttr.pt(1);
	const Vector2d  pt2 = ttr.pt(2);
	const Vector2d  pt4 = ttr.pt(4); Vector2d dpt;
	Rectangle3d _rct_;
	_rct_.Origin() = Vector3d(pt1.X(), pt1.Y(), 0.0);
	dpt = pt2 - pt1; _rct_.Edge0 () = Vector3d(dpt.X(), dpt.Y(), 0.0);
	dpt = pt4 - pt1; _rct_.Edge1 () = Vector3d(dpt.X(), dpt.Y(), 0.0);

	const double dist = Distance(_Ctr_, _Fr0_);
	const double Dist = Distance(_Ctr_, _rct_);
	const double corr = 1 - dist/crt();

	const bool nearblg = (corr > 0.0);
	// const bool outside = (Dist >= 1.0e-3);
	const bool outside = (Dist > 0.0);
	const bool infront = (fabs(dist - Dist) <= 1.0e-2);

	if(nearblg && outside && infront)
	{
		const double _h_s = _ctr_.Z();
		const double _h_b = hbl();
		const double _w_1 = wb1();
		const double _w_2 = wb2();
		const double nond = _h_s/_h_b;

		double fct;
		if(typ() == 1)
		{
			if(_w_1 <= _h_b)
			{
				if(false){/* dummy */}
				else if(nond <= 1.0){fct = 2.0;}
				else if(nond <= 1.5){fct = 6.0 - 4.0*nond;}
				else                {fct = 0.0;}
			}
			else
			{
				if(false){/* dummy */}
				else if(nond <= 1.0 ){fct = 2.0;}
				else if(nond <= 1.67){fct = 5.0 - 3.0*nond;}
				else                 {fct = 0.0;}
			}
		}
		else
		{
			if(false){/* dummy */}
			else if(nond <= 1.0 ){fct = 2.5;}
			else if(nond <= 1.83){fct = 5.5 - 3.0*nond;}
			else                 {fct = 0.0;}
		}

		const double w_p = corr*scl(); const double dys = fct*w_p*smt(); if(fabs(dys) < 1.0e-3) return;



		// Delta y_s evaluation done. Final task is evaluation of the translation vector.

		const Vector2d transl = - dist*_nrm_;
		const Vector2d _fr1__ = _fr1_  + transl;
		const Vector2d _dir1_ = _fr1__ - Vector2d(_ctr_.X(), _ctr_.Y());
		const Vector2d _fr2__ = _fr2_  + transl;
		const Vector2d _dir2_ = _fr2__ - Vector2d(_ctr_.X(), _ctr_.Y());
		const double len1 = _dir1_.Length();
		const double len2 = _dir2_.Length();

		if(_dir1_.Dot(_dir2_) > 0 || len1 < 1.0e-3 || len2 < 1.0e-3) return;

		const Vector2d _dir__ = (_dir1_.Dot(e1) > 0 ? _dir1_ : _dir2_);
		const Vector2d _nrm__ = (_dir__.Dot(e2) > 0 ?     e2 :   - e2);

		const double  proj = _dir__.Dot(_nrm__);
		const Vector2d res = (dys <= proj ? _dir__*dys/proj : _dir__ + (dys - proj)*_nrm__);
		d_x_ = res.X(); d_y_ = res.Y();
	}
}

void BuildingParameters::dump(int w_d)
{
	ML_TRACE("BuildingParameters::dump\n");

	// Placeholders for source object
	Ring _rin_; bool _stk_, _plm_; double _emi_,_vel_,_gas_,_tpr_;
	ptr_src->getPointSrc(_rin_, _stk_, _plm_, _emi_, _vel_, _gas_, _tpr_);

	// Placeholders for ring object
	int  _ide_; Vector3d _ctr_; double _dia_,_hhh_; string _des_;
	_rin_.getRing(_ide_, _ctr_, _dia_, _hhh_, _des_);

	const int ide = _ide_;

	os << setw(COLUMN) << ide << ',';
	os << setw(COLUMN) << w_d << ',';
	os << setw(COLUMN) << w_1 << ',';
	os << setw(COLUMN) << w_2 << ',';
	os << setw(COLUMN) << h_h << ',';
	os << setw(COLUMN) << t_p << ',';
	os << setw(COLUMN) << h_e << ',';
	os << setw(COLUMN) << d_x << ',';
	os << setw(COLUMN) << d_y << ',';
	os << setw(COLUMN) << z_1 << ',';
	os << setw(COLUMN) << z_2 << ',';
	os << setw(COLUMN) << y_1 << ',';
	os << setw(COLUMN) << y_2 << ',';
	os << setw(COLUMN) << a_l << ',';
	os << setw(COLUMN) << e_f << endl;

	CLST << setw(COLUMN) << ide;
	CLST << setw(COLUMN) << w_d;
	CLST << setw(COLUMN) << w_1;
	CLST << setw(COLUMN) << w_2;
	CLST << setw(COLUMN) << h_h;
	CLST << setw(COLUMN) << t_p;
	CLST << setw(COLUMN) << h_e;
	CLST << setw(COLUMN) << d_x;
	CLST << setw(COLUMN) << d_y;
	CLST << setw(COLUMN) << z_1;
	CLST << setw(COLUMN) << z_2;
	CLST << setw(COLUMN) << y_1;
	CLST << setw(COLUMN) << y_2;
	CLST << setw(COLUMN) << a_l;
	CLST << setw(COLUMN) << e_f;
	CLST << setw(COLUMN) << (e_f == 0 ? 0 : getTheta(true));
	CLST << setw(COLUMN) << getMaxGEP() << endl;
}
