#include "NR.h"
#include "ErrorHandler.h"
#include "../share/common.h"

// The following codes can be obtained from Numerical Recipes in C.
// Some modifications had been done for converting to C++, simplifications or extensions.

double NR::SpecialFunctions::gammln(double xx)
{
	double x,y,tmp,ser;
	static double cof[6]={76.18009172947146,-86.50532032941677,
		24.01409824083091,-1.231739572450155,
		0.1208650973866179e-2,-0.5395239384953e-5};
	int j;
	
	y=x=xx;
	tmp=x+5.5;
	tmp -= (x+0.5)*log(tmp);
	ser=1.000000000190015;
	for (j=0;j<=5;j++) ser += cof[j]/++y;
	return -tmp+log(2.5066282746310005*ser/x);
}


double NR::SpecialFunctions::beta(double z, double w)
{
	return exp(gammln(z)+gammln(w)-gammln(z+w));
}


#define ITMAX 100
#define EPS 3.0e-7
void NR::SpecialFunctions::gser(double *gamser, double a, double x, double *gln)
{
	int n;
	double sum,del,ap;
	
	try{
		*gln=gammln(a);
		if (x <= 0.0) {
			if (x < 0.0) throw 1;
			*gamser=0.0;
			return;
		} else {
			ap=a;
			del=sum=1.0/a;
			for (n=1;n<=ITMAX;n++) {
				++ap;
				del *= x/ap;
				sum += del;
				if (fabs(del) < fabs(sum)*EPS) {
					*gamser=sum*exp(-x+a*log(x)-(*gln));
					return;
				}
			}
			throw 2;
			return;
		}
	}
	catch(int checkpoint)
	{
		switch(checkpoint)
		{
		case 1:
			cerr << ErrorNR_gser_1(x);
			CERR << ErrorNR_gser_1(x);
			break;
		case 2:
			cerr << ErrorNR_gser_2(a, ITMAX);
			CERR << ErrorNR_gser_2(a, ITMAX);
			break;
		}
		exit(1);
	}
}
#undef ITMAX
#undef EPS


#define ITMAX 100
#define EPS 3.0e-7
#define FPMIN 1.0e-30
void NR::SpecialFunctions::gcf(double *gammcf, double a, double x, double *gln)
{
	int i;
	double an,b,c,d,del,h;
	
	try{
		*gln=gammln(a);
		b=x+1.0-a;
		c=1.0/FPMIN;
		d=1.0/b;
		h=d;
		for (i=1;i<=ITMAX;i++) {
			an = -i*(i-a);
			b += 2.0;
			d=an*d+b;
			if (fabs(d) < FPMIN) d=FPMIN;
			c=b+an/c;
			if (fabs(c) < FPMIN) c=FPMIN;
			d=1.0/d;
			del=d*c;
			h *= del;
			if (fabs(del-1.0) < EPS) break;
		}
		if (i > ITMAX) throw 1;
		*gammcf=exp(-x+a*log(x)-(*gln))*h;
	}
	catch(int checkpoint)
	{
		switch(checkpoint)
		{
		case 1:
			cerr << ErrorNR_gcf(a, ITMAX);
			CERR << ErrorNR_gcf(a, ITMAX);
			break;
		}
		exit(1);
	}
}
#undef ITMAX
#undef EPS
#undef FPMIN


double NR::SpecialFunctions::gammp(double a, double x)
{
	double gamser,gammcf,gln;
	
	try{
		if (x < 0.0 || a <= 0.0) throw 1;
	
		if (x < (a+1.0)) {
			gser(&gamser,a,x,&gln);
			return gamser;
		} else {
			gcf(&gammcf,a,x,&gln);
			return 1.0-gammcf;
		}
	}
	catch(int checkpoint)
	{
		switch(checkpoint)
		{
		case 1:
			cerr << ErrorNR_gammp(a, x);
			CERR << ErrorNR_gammp(a, x);
			break;
		}
		exit(1);
	}
}


double NR::SpecialFunctions::gammq(double a, double x)
{
	double gamser,gammcf,gln;
	
	try{
		if (x < 0.0 || a <= 0.0) throw 1;
		if (x < (a+1.0)) {
			gser(&gamser,a,x,&gln);
			return 1.0-gamser;
		} else {
			gcf(&gammcf,a,x,&gln);
			return gammcf;
		}
	}
	catch(int checkpoint)
	{
		switch(checkpoint)
		{
		case 1:
			cerr << ErrorNR_gammq(a, x);
			CERR << ErrorNR_gammq(a, x);
			break;
		}
		exit(1);
	}
}


double NR::SpecialFunctions::erff(double x)
{
	return x < 0.0 ? -gammp(0.5,x*x) : gammp(0.5,x*x);
}


double NR::SpecialFunctions::erffc(double x){
	return x < 0.0 ? 1.0+gammp(0.5,x*x) : gammq(0.5,x*x);
}


double NR::SpecialFunctions::erfcc(double x)
{
	double t,z,ans;
	
	z=fabs(x);
	t=1.0/(1.0+0.5*z);
	ans=t*exp(-z*z-1.26551223+t*(1.00002368+t*(0.37409196+t*(0.09678418+
		t*(-0.18628806+t*(0.27886807+t*(-1.13520398+t*(1.48851587+
		t*(-0.82215223+t*0.17087277)))))))));
	return x >= 0.0 ? ans : 2.0-ans;
}


double NR::SpecialFunctions::chebev(double a, double b, double c[], int m, double x)
{
	double d=0.0,dd=0.0,sv,y,y2;
	int j;
	
	try{
		if ((x-a)*(x-b) > 0.0) throw 1;
		y2=2.0*(y=(2.0*x-a-b)/(b-a));
		for (j=m-1;j>=1;j--) {
			sv=d;
			d=y2*d-dd+c[j];
			dd=sv;
		}
		return y*d-dd+0.5*c[0];
	}
	catch(int checkpoint)
	{
		switch(checkpoint)
		{
		case 1:
			cerr << ErrorNR_chebev(a, b, x);
			CERR << ErrorNR_chebev(a, b, x);
			break;
		}
		exit(1);
	}
}


#define NUSE1 7
#define NUSE2 7
void NR::SpecialFunctions::beschb(double x, double *gam1, double *gam2, double *gampl, double *gammi)
{
	double xx;
	static double c1[] = {
		-1.142022680371168e0,6.5165112670737e-3,
		3.087090173086e-4,-3.4706269649e-6,6.9437664e-9,
		3.67795e-11,-1.356e-13};
	static double c2[] = {
		1.843740587300905e0,-7.68528408447867e-2,
		1.2719271366546e-3,-4.9717367042e-6,-3.31261198e-8,
		2.423096e-10,-1.702e-13,-1.49e-15};
		
	xx=8.0*x*x-1.0;
	*gam1=chebev(-1.0,1.0,c1,NUSE1,xx);
	*gam2=chebev(-1.0,1.0,c2,NUSE2,xx);
	*gampl= *gam2-x*(*gam1);
	*gammi= *gam2+x*(*gam1);
}
#undef NUSE1
#undef NUSE2


#define EPS 1.0e-10
#define FPMIN 1.0e-30
#define MAXIT 10000
#define XMIN 2.0
#define PI 3.141592653589793
void NR::SpecialFunctions::bessik(double x, double xnu, double *ri, double *rk, double *rip, double *rkp)
{
	int i,l,nl;
	double a,a1,b,c,d,del,del1,delh,dels,e,f,fact,fact2,ff,gam1,gam2,
		gammi,gampl,h,p,pimu,q,q1,q2,qnew,ril,ril1,rimu,rip1,ripl,
		ritemp,rk1,rkmu,rkmup,rktemp,s,sum,sum1,x2,xi,xi2,xmu,xmu2;

	try{
		if (x <= 0.0 || xnu < 0.0) throw 1;
		nl=(int)(xnu+0.5);
		xmu=xnu-nl;
		xmu2=xmu*xmu;
		xi=1.0/x;
		xi2=2.0*xi;
		h=xnu*xi;
		if (h < FPMIN) h=FPMIN;
		b=xi2*xnu;
		d=0.0;
		c=h;
		for (i=1;i<=MAXIT;i++) {
			b += xi2;
			d=1.0/(b+d);
			c=b+1.0/c;
			del=c*d;
			h=del*h;
			if (fabs(del-1.0) < EPS) break;
		}
		if (i > MAXIT) throw 2;
		ril=FPMIN;
		ripl=h*ril;
		ril1=ril;
		rip1=ripl;
		fact=xnu*xi;
		for (l=nl;l>=1;l--) {
			ritemp=fact*ril+ripl;
			fact -= xi;
			ripl=fact*ritemp+ril;
			ril=ritemp;
		}
		f=ripl/ril;
		if (x < XMIN) {
			x2=0.5*x;
			pimu=PI*xmu;
			fact = (fabs(pimu) < EPS ? 1.0 : pimu/sin(pimu));
			d = -log(x2);
			e=xmu*d;
			fact2 = (fabs(e) < EPS ? 1.0 : sinh(e)/e);
			beschb(xmu,&gam1,&gam2,&gampl,&gammi);
			ff=fact*(gam1*cosh(e)+gam2*fact2*d);
			sum=ff;
			e=exp(e);
			p=0.5*e/gampl;
			q=0.5/(e*gammi);
			c=1.0;
			d=x2*x2;
			sum1=p;
			for (i=1;i<=MAXIT;i++) {
				ff=(i*ff+p+q)/(i*i-xmu2);
				c *= (d/i);
				p /= (i-xmu);
				q /= (i+xmu);
				del=c*ff;
				sum += del;
				del1=c*(p-i*ff);
				sum1 += del1;
				if (fabs(del) < fabs(sum)*EPS) break;
			}
			if (i > MAXIT) throw 3;
			rkmu=sum;
			rk1=sum1*xi2;
		} else {
			b=2.0*(1.0+x);
			d=1.0/b;
			h=delh=d;
			q1=0.0;
			q2=1.0;
			a1=0.25-xmu2;
			q=c=a1;
			a = -a1;
			s=1.0+q*delh;
			for (i=2;i<=MAXIT;i++) {
				a -= 2*(i-1);
				c = -a*c/i;
				qnew=(q1-b*q2)/a;
				q1=q2;
				q2=qnew;
				q += c*qnew;
				b += 2.0;
				d=1.0/(b+a*d);
				delh=(b*d-1.0)*delh;
				h += delh;
				dels=q*delh;
				s += dels;
				if (fabs(dels/s) < EPS) break;
			}
			if (i > MAXIT) throw 4;
			h=a1*h;
			rkmu=sqrt(PI/(2.0*x))*exp(-x)/s;
			rk1=rkmu*(xmu+x+0.5-h)*xi;
		}
		rkmup=xmu*xi*rkmu-rk1;
		rimu=xi/(f*rkmu-rkmup);
		*ri=(rimu*ril1)/ril;
		*rip=(rimu*rip1)/ril;
		for (i=1;i<=nl;i++) {
			rktemp=(xmu+i)*xi2*rk1+rkmu;
			rkmu=rk1;
			rk1=rktemp;
		}
		*rk=rkmu;
		*rkp=xnu*xi*rkmu-rk1;
	}
	catch(int checkpoint)
	{
		switch(checkpoint)
		{
		case 1:
			cerr << ErrorNR_bessik_1(xnu, x);
			CERR << ErrorNR_bessik_1(xnu, x);
			break;
		case 2:
			cerr << ErrorNR_bessik_2(x);
			CERR << ErrorNR_bessik_2(x);
			break;
		case 3:
			cerr << ErrorNR_bessik_3(MAXIT);
			CERR << ErrorNR_bessik_3(MAXIT);
			break;
		case 4:
			cerr << ErrorNR_bessik_4(MAXIT);
			CERR << ErrorNR_bessik_4(MAXIT);
			break;
		}
		exit(1);
	}
}
#undef EPS
#undef FPMIN
#undef MAXIT
#undef XMIN
#undef PI


#define PI 3.141592653589793
void NR::SpecialFunctions::bessIK(double x, double xnu, double *ri, double *rk, double *rip, double *rkp){
	if(xnu >= 0)
		bessik(x, xnu, ri, rk, rip, rkp);
	else{
		double nu = - xnu;
		bessik(x,  nu, ri, rk, rip, rkp);
		*ri = *ri + 2*(*rk)*sin(nu*PI)/PI;	// Used the reflection formula
	}
}
#undef PI
