#include "simulator.h"
#include "ErrorHandler.h"
#include "../share/common.h"
#include "../share/WildMagic.h"


Simulator::Simulator()
{
	scan_met = scan_rcp = scan_src = print_met = print_rcp = print_src = true;
	ptr_met = NULL;
	ptr_src = NULL;
	ptr_rcp = NULL;
	ptr_sgm = NULL;
	ptr_pat = NULL;
	ptr_ran = NULL;
	ptr_blg = NULL;
	ptr_adj = NULL;
	virtual_latr = virtual_vert = 0;
	sz0  = sy0  = 0;
	is_day = is_calm = false;
}

Simulator::~Simulator()
{
	ML_TRACE("Simulator::~Simulator");

	if(ptr_rcp->getTerrainOpt()) detachAdjData();

	detachRcpData();
	detachMetData();
	detachSrcData();
	detachSgmData();
	detachPatData();
	detachRanData();
	detachBlgData();

	if(isTimeSeries())
	{
		delete [] lt_res;
		delete [] lt_dep;
		delete [] lt_flg;
		delete [] lt_nrs;
		delete tim;
	}
	else
	{
		delete [] st_res;
		delete [] st_dep;
		delete [] st_flg;
	}
}

/*
*		Definition of fixed coordinate used in this simulator:
*
*		Y=North
*		|
*		|
*		|
*		|------->X=East
*/

double Simulator::getWindDirectionALaMathematician(double blowing_direction_measured_from_north_with_clockwise_in_degrees) const
{
	ML_TRACE("Simulator::getWindDirectionALaMathematician");

	const double deg = blowing_direction_measured_from_north_with_clockwise_in_degrees;
	return DEG2RAD*(deg <= 90 ? 90 - deg : 450 - deg);
}

void Simulator::getLocalCoor(double& x_loc, double& y_loc, double x_glo, double y_glo) const
{
	ML_TRACE("Simulator::getLocalCoor");

	// Placeholders for receptor coordinates
	double _x_,_y_,_z_,_h_;

	const double  rad = getWindDirectionALaMathematician(getModelled_wd());
	ptr_rcp->getCoor(_x_, _y_, _z_,_h_);
	const Vector2d org = Vector2d(_x_, _y_);
	const Vector2d glo = Vector2d(x_glo, y_glo);
	const Vector2d dir = Vector2d(cos(rad), sin(rad));
	const Vector2d crs = Vector2d(-dir.Y(), dir.X());
	x_loc = (glo - org).Dot(dir);
	y_loc = (glo - org).Dot(crs);
}

void Simulator::scan()
{
	ML_TRACE("Simulator::scan");

	if(scan_rcp) ptr_rcp->scan();
	if(scan_src) ptr_src->scan();
	if(scan_met)
	{
		if(isTimeSeries())
		{
			ptr_met->scan(); tim = new TimeCode(); tim->scan();

			const int n1 = tim-> getMaxDailyCodes();
			const int n2 = tim->getMaxHourlyCodes();
			const int n3 = ptr_rcp->getNoElms();

			lt_res = new double [(n1+1)*(n2+1)*n3];
			lt_dep = new double [(n1+1)*(n2+1)*n3];
			lt_flg = new double [(n1+1)*(n2+1)*n3];
			lt_nrs = new int    [(n1+1)*(n2+1)];

			int i;
			for(i = 0; i < (n1+1)*(n2+1)*n3; i++)
				lt_res[i] = 0;
			for(i = 0; i < (n1+1)*(n2+1)*n3; i++)
				lt_dep[i] = 0;
			for(i = 0; i < (n1+1)*(n2+1)*n3; i++)
				lt_flg[i] = 1;
			for(i = 0; i < (n1+1)*(n2+1); i++)
				lt_nrs[i] = 0;

			ptr_sgm->setDilutionFactor(1.0);
		}
		else
		{
 			ptr_met->scan();
 			const double avg_sec = ptr_met->getAverageSeconds();
 			const double dilution = pow(avg_sec/180.0, ptr_met->getPower4AverageSeconds());
			ptr_sgm->setDilutionFactor(dilution);

			const int n3 = ptr_rcp->getNoElms();

			st_res = new double [n3];
			st_dep = new double [n3];
			st_flg = new double [n3];

			int i;
			for(i = 0; i < n3; i++)
				st_res[i] = 0;
			for(i = 0; i < n3; i++)
				st_dep[i] = 0;
			for(i = 0; i < n3; i++)
				st_flg[i] = 1;
 		}
	}

	ifstream is("../tmp/runtime.in.csv");

	is.exceptions(ios_base::goodbit);
	try{
		is.exceptions(ios_base::badbit | ios_base::failbit);

		int int_placeholder; double dbl_placeholder; string str_placeholder;

		is >> int_placeholder; setProgressPeekStep(int_placeholder);
		is.ignore(NO_OF_CHARS_PER_ONE_LINE, ',');
		getline(is, str_placeholder, ','); setPrefix(str_placeholder);

		is >> int_placeholder;
		setConcentrationUnitFlag(int_placeholder == 1);
		ptr_rcp->setConcentrationUnitFlag(int_placeholder == 1);
		is.ignore(NO_OF_CHARS_PER_ONE_LINE, ',');
		is.ignore(NO_OF_CHARS_PER_ONE_LINE, ',');

		is >> int_placeholder; setMinimumIterations(int_placeholder);
		is.ignore(NO_OF_CHARS_PER_ONE_LINE, ',');
		is >> int_placeholder; setMaximumIterations(int_placeholder);
		is.ignore(NO_OF_CHARS_PER_ONE_LINE, ',');
		is >> dbl_placeholder; setToleranceRatio(dbl_placeholder);

		is >> int_placeholder; setListLevel(int_placeholder);
		is.ignore(NO_OF_CHARS_PER_ONE_LINE, ',');
		is.ignore(NO_OF_CHARS_PER_ONE_LINE, ',');

		is.close();
	}
	catch(...)
	{
		cerr << ErrorSimulator_scan("runtime.in.csv");
		CERR << ErrorSimulator_scan("runtime.in.csv");
		exit(1);
	}
}

void Simulator::print()
{
#if SCAN_CHECK > 0
	if(print_src) ptr_src->print();
	if(print_met) ptr_met->print();
	if(print_rcp) ptr_rcp->print();
#endif
}

void Simulator::lists()
{
	ML_TRACE("Simulator::lists");

	const ListLevel level = getListLevel();

	if(print_src && (level == MODERATE || level == FULL)) ptr_src->lists();
	if(print_met && (level == MODERATE || level == FULL)) ptr_met->lists();
	if(print_rcp && (level == FULL))                      ptr_rcp->lists();
}

double Simulator::getTrigHalfOpenAngle(Trig tr)
{
	ML_TRACE("Simulator::getTrigHalfOpenAngle");

	switch(tr)
	{
	case COS:
		return cos(M_PI/360*getOpenAngle());
	case SIN:
		return sin(M_PI/360*getOpenAngle());
	case TAN:
		return tan(M_PI/360*getOpenAngle());
	default:
		throw;
	}
}

void Simulator::assembleResult(int counter)
{
	ML_TRACE("Simulator::assembleResult");

	if(isTimeSeries())
	{
		// Placeholders for retrieving the current time
		int _yr, _mn, _dy, _hr;
		ptr_met->getTime(_yr, _mn, _dy, _hr);

		tim->setTime(_yr, _mn, _dy, _hr);
		const int nr_elms = ptr_rcp->getNoElms();
		const int dy_code = tim-> getDailyCode();
		const int hr_code = tim->getHourlyCode();
		const int nr_dy_codes_p = tim-> getMaxDailyCodes() + 1;
		const int nr_hr_codes_p = tim->getMaxHourlyCodes() + 1;

		const int nrs_d_h = (hr_code-1)       + (dy_code-1)      *nr_hr_codes_p;
		const int nrs_d_0 = (nr_hr_codes_p-1) + (dy_code-1)      *nr_hr_codes_p;
		const int nrs_0_h = (hr_code-1)       + (nr_dy_codes_p-1)*nr_hr_codes_p;
		const int nrs_0_0 = (nr_hr_codes_p-1) + (nr_dy_codes_p-1)*nr_hr_codes_p;
		lt_nrs[nrs_d_h]  += 1;
		lt_nrs[nrs_d_0]  += 1;
		lt_nrs[nrs_0_h]  += 1;
		lt_nrs[nrs_0_0]  += 1;

		ptr_rcp->head();
		for(int i = 1; i <= nr_elms; i++)
		{
			const int res_d_h = (i-1) + (hr_code-1)      *nr_elms + (dy_code-1)      *nr_elms*nr_hr_codes_p;
			const int res_d_0 = (i-1) + (nr_hr_codes_p-1)*nr_elms + (dy_code-1)      *nr_elms*nr_hr_codes_p;
			const int res_0_h = (i-1) + (hr_code-1)      *nr_elms + (nr_dy_codes_p-1)*nr_elms*nr_hr_codes_p;
			const int res_0_0 = (i-1) + (nr_hr_codes_p-1)*nr_elms + (nr_dy_codes_p-1)*nr_elms*nr_hr_codes_p;
			lt_res[res_d_h]  += ptr_rcp->valuePt(Rcp::CON);
			lt_res[res_d_0]  += ptr_rcp->valuePt(Rcp::CON);
			lt_res[res_0_h]  += ptr_rcp->valuePt(Rcp::CON);
			lt_res[res_0_0]  += ptr_rcp->valuePt(Rcp::CON);
			lt_dep[res_d_h]  += ptr_rcp->valuePt(Rcp::DEP);
			lt_dep[res_d_0]  += ptr_rcp->valuePt(Rcp::DEP);
			lt_dep[res_0_h]  += ptr_rcp->valuePt(Rcp::DEP);
			lt_dep[res_0_0]  += ptr_rcp->valuePt(Rcp::DEP);
			lt_flg[res_d_h]  *= ptr_rcp->valuePt(Rcp::FLG);
			lt_flg[res_d_0]  *= ptr_rcp->valuePt(Rcp::FLG);
			lt_flg[res_0_h]  *= ptr_rcp->valuePt(Rcp::FLG);
			lt_flg[res_0_0]  *= ptr_rcp->valuePt(Rcp::FLG);
			ptr_rcp->advance();
		}
		ptr_rcp->head();
	}
	else
	{
		const int nr_elms = ptr_rcp->getNoElms();

		ptr_rcp->head();
		for(int i = 1; i <= nr_elms; i++)
		{
			const int res_idx = i-1;
			st_res[res_idx]  += ptr_rcp->valuePt(Rcp::CON);
			st_dep[res_idx]  += ptr_rcp->valuePt(Rcp::DEP);
			st_flg[res_idx]  *= ptr_rcp->valuePt(Rcp::FLG);
			ptr_rcp->advance();
		}
		ptr_rcp->head();

		const string old = ptr_rcp->getOsFile();

		char buffer[8]; _itoa(counter, buffer, 10); string filename = buffer;

		if(false){/* dummy */}
		else if(counter < 10)
			filename = "000" + filename;
		else if(counter < 100)
			filename =  "00" + filename;
		else if(counter < 1000)
			filename =   "0" + filename;
		filename = "../tmp/" + getPrefix() + "_" + filename + ".out.csv";

		ptr_rcp->setOsFile(filename); ptr_rcp->print(true); const ListLevel level = getListLevel();
		ptr_rcp->lists(counter, (level == MODERATE || level == FULL));
		ptr_rcp->setOsFile(old);

		if(counter == ptr_met->getNoElms())
		{
			ptr_rcp->head();
			for(int j = 1; j <= nr_elms; j++)
			{
				const int res_idx = j-1;
				ptr_rcp->valuePt(Rcp::CON) = st_res[res_idx]/ptr_met->getNoElms();
				ptr_rcp->valuePt(Rcp::DEP) = st_dep[res_idx]/ptr_met->getNoElms();
				ptr_rcp->valuePt(Rcp::FLG) = st_flg[res_idx];
				ptr_rcp->advance();
			}
			ptr_rcp->head();

			filename = "../tmp/" + getPrefix() + "_" + "0000.out.csv";
			ptr_rcp->setOsFile(filename); ptr_rcp->print(true);
			ptr_rcp->lists(0, (level == MODERATE || level == FULL));
			ptr_rcp->setOsFile(old);
		}
	}
}

int Simulator::reclassST(int st)
{
	ML_TRACE("Simulator::reclassST");

	if(isTimeSeries())
	{
		if(false){/* dummy line*/ }     				//  before  after
		else if(st <=  2){setDayFlag(true) ; return 1;}	//    A,AB->A
		else if(st <=  4){setDayFlag(true) ; return 2;}	//    B,BC->B
		else if(st <=  6){setDayFlag(true) ; return 3;}	//    C,CD->C
		else if(st <=  7){setDayFlag(true) ; return 4;}	//      DD->DD
		else if(st <=  8){setDayFlag(false); return 4;}	//      DN->DN
		else if(st ==  9){setDayFlag(false); return 5;}	//       E->E
		else if(st <= 11){setDayFlag(false); return 6;}	//     F,G->F
		else             {setDayFlag(false); return 9999;}
	}
	else
	{
		if(st >= 5)
		{
			setDayFlag(false); return st - 1;
		}
		else
		{
			setDayFlag(true) ; return st;
		}
	}
}

void Simulator::addSrcContrib(double corr_factor)
{
	ML_TRACE("Simulator::addSrcContrib");

	ptr_rcp->valuePt(Rcp::CON) += concentration*corr_factor;
	ptr_rcp->valuePt(Rcp::DEP) += deposition   *corr_factor;
	ptr_rcp->valuePt(Rcp::FLG) *= validation;
}

void Simulator::setResult(Rcp::Result res, double val)
{
	ML_TRACE("Simulator::setResult");

	switch(res)
	{
	case Rcp::CON:
		concentration = (getConcentrationUnitFlag() ? val/ptr_src->getMolecularWeight()*22.4*288.15/273.15 : val);
		break;
	case Rcp::DEP:
		deposition    = val;
		break;
	case Rcp::FLG:
		validation    = val;
		break;
	default:
		throw;
	}
}

double Simulator::getResult(Rcp::Result res) const
{
	ML_TRACE("Simulator::getResult");

	switch(res)
	{
	case Rcp::CON:
		return concentration;
	case Rcp::DEP:
		return deposition;
	case Rcp::FLG:
		return validation;
	default:
		throw;
	}
}

bool Simulator::isBadData() const
{
	ML_TRACE("Simulator::isBadData");

	// Placeholders for meteorological data
	double _wd_, _ws_, _tp_; int _st_; bool bad = false;

	ptr_met->getMet(_wd_, _ws_, _st_, _tp_);

	if(_ws_ < 0.0 || _ws_ > 999.0) bad = true;
	if(_st_ < 1   || _st_ > 11   ) bad = true;
	if(_tp_ < 0.0 || _tp_ > 999.0) bad = true;

	if(bad)
	{
		char buffer[50];
		_gcvt(_ws_, 8,  buffer); string str_ws = buffer;
		_gcvt(_tp_, 8,  buffer); string str_tp = buffer;
		_itoa(_st_, buffer, 10); string str_st = buffer;

		string msg = "";

		msg += "\n";
		msg += "+WARNING MESSAGE\n";
		if(isTimeSeries())
		{
			// Placeholders for retrieving the current time
			int _yr, _mn, _dy, _hr;
			ptr_met->getTime(_yr, _mn, _dy, _hr);

			_itoa(_yr, buffer, 10); string str_yr = buffer;
			_itoa(_mn, buffer, 10); string str_mn = buffer;
			_itoa(_dy, buffer, 10); string str_dy = buffer;
			_itoa(_hr, buffer, 10); string str_hr = buffer;

			msg += " - Bad input data detected at " + str_yr + " / " +  str_mn + " / " + str_dy + " - " + str_hr + ":" + "00.\n";
		}
		else
		{
			msg += " - Bad input data detected at this case.\n";
		}
		msg += "+STATUS\n";
		msg += " - Ignored.\n";
		msg += "+INPUT DATA\n";
		msg += " - 1) Wind speed  -> " + str_ws + " (m/s).\n";
		msg += " - 2) Stability   -> " + str_st + ".\n";
		msg += " - 3) Temperature -> " + str_tp + " (K).\n";
		msg += "+RESOLUTIONS\n";
		msg += " - First of all it is recommended to check your input meteorological data file.\n";
		msg += " - If the required data is not available, you may safely ignore this evaluation.\n";
		msg += "Process continued...\n\n";
		CLOG << msg;
	}

	return bad;
}

double Simulator::power4ws(int st_, UrbanOrRural uor) const
{
	ML_TRACE("Simulator::power4ws");

	try{
		switch(uor)
		{
		case RURAL:
			switch(st_)
			{
			case 1:
				return 0.07;
			case 2:
				return 0.07;
			case 3:
				return 0.10;
			case 4:
				return 0.15;
			case 5:
				return 0.35;
			case 6:
				return 0.55;
			default:
				throw 1;
			}
		case URBAN:
			switch(st_)
			{
			case 1:
				return 0.15;
			case 2:
				return 0.15;
			case 3:
				return 0.20;
			case 4:
				return 0.25;
			case 5:
				return 0.30;
			case 6:
				return 0.30;
			default:
				throw 1;
			}
		default:
			cout << "Simulator::power4ws::Unrecognized enum type value detected." << endl;
			exit(1);
		}
	}
	catch(int checkpoint)
	{
		switch(checkpoint)
		{
		case 1:
			cerr << ErrorSimulator_power4ws(st_);
			CERR << ErrorSimulator_power4ws(st_);
			break;
		}
		exit(1);
	}
}

void Simulator::setListLevel(int list_level_)
{
	ML_TRACE("Simulator::setListLevel");

	switch(list_level_)
	{
	case 1:
		list_level = ESSENTIAL;
		break;
	case 2:
		list_level = MODERATE;
		break;
	case 3:
		list_level = FULL;
		break;
	default:
		cout << "Simulator::setListLevel::Unrecognized outlist level number detected." << endl;
		exit(1);
	}
}
